package cn.kdan.compdf.service.impl;

import cn.kdan.compdf.config.ExcelWidthStyleStrategy;
import cn.kdan.compdf.config.RedisConstantServer;
import cn.kdan.compdf.dto.DashboardStatisticsDTO;
import cn.kdan.compdf.entity.BackgroundConvertData;
import cn.kdan.compdf.entity.dto.AnalysisDataDTO;
import cn.kdan.compdf.entity.dto.DashboardDetailedDataDTO;
import cn.kdan.compdf.entity.dto.DashboardExportObject;
import cn.kdan.compdf.entity.dto.DashboardQueryDTO;
import cn.kdan.compdf.enums.BackgroundErrorEnum;
import cn.kdan.compdf.enums.ResponseEnum;
import cn.kdan.compdf.exception.CommonException;
import cn.kdan.compdf.mapper.BackgroundConvertDataMapper;
import cn.kdan.compdf.service.BackgroundConvertDataService;
import cn.kdan.compdf.service.BackgroundUserService;
import com.alibaba.excel.EasyExcel;
import com.alibaba.excel.util.MapUtils;
import com.alibaba.excel.write.style.column.LongestMatchColumnWidthStyleStrategy;
import com.alibaba.fastjson.JSON;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import lombok.extern.slf4j.Slf4j;
import org.springframework.cache.CacheManager;
import org.springframework.cache.annotation.CacheEvict;
import org.springframework.cache.annotation.Cacheable;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.stereotype.Service;

import javax.annotation.Resource;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.math.BigDecimal;
import java.net.URLEncoder;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.temporal.ChronoUnit;
import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

/**
 * @author comPDF-Kit WPH 2022-10-09
 */
@Service
@Slf4j
public class BackgroundConvertDataServiceImpl extends ServiceImpl<BackgroundConvertDataMapper, BackgroundConvertData> implements BackgroundConvertDataService {

    @Resource(name = "myCacheManager")
    private CacheManager cacheManager;

    private final StringRedisTemplate redisTemplate;

    private final BackgroundUserService userService;

    public BackgroundConvertDataServiceImpl(StringRedisTemplate redisTemplate, BackgroundUserService userService) {
        this.redisTemplate = redisTemplate;
        this.userService = userService;
    }

    @Override
    public IPage<BackgroundConvertData> page(BackgroundConvertData query) {
        return this.page(new Page<>(query.getPage(), query.getSize()), Wrappers.query(query));
    }

    @Override
    @Cacheable(value = RedisConstantServer.DASHBOARD_ANALYSIS_DATA, cacheManager = "myCacheManager", key = "T(String).valueOf(#queryDTO.toString())", unless = "#result == null")
    public AnalysisDataDTO getAnalysisData(DashboardQueryDTO queryDTO) {
        // 获取一个查询query
        LambdaQueryWrapper<BackgroundConvertData> queryWrapper = getDashboardQueryWrapper(queryDTO);

        List<BackgroundConvertData> backgroundConvertData = this.baseMapper.selectList(queryWrapper);
        if (backgroundConvertData == null || backgroundConvertData.size() <= 0) {
            return null;
        }
        long fileTotal = 0; // 文件总数
        long successfulRequest = 0; // 成功数
        long errorRequest = 0;  // 失败数
        double errorRatio = 0;  // 失败率
        double averageProcessTime = 0;  // 平均处理时长
        BigDecimal averageProcessTimeTotal = BigDecimal.ZERO; // 总处理时长
        for (BackgroundConvertData backgroundConvertDatum : backgroundConvertData) {
            fileTotal += backgroundConvertDatum.getFileTotal();
            successfulRequest += backgroundConvertDatum.getFileSuccessTotal();
            errorRequest += backgroundConvertDatum.getFileFailedTotal();
            averageProcessTimeTotal = backgroundConvertDatum.getFileConvertTime().add(averageProcessTimeTotal);
        }
        BigDecimal averageProcessTime1 = averageProcessTimeTotal.divide(new BigDecimal(backgroundConvertData.size()), 2, BigDecimal.ROUND_HALF_UP);
        return AnalysisDataDTO.builder()
                .fileTotal(fileTotal)
                .successfulRequest(successfulRequest)
                .errorRequest(errorRequest)
                .errorRatio(new BigDecimal(String.format("%.2f", errorRequest * 100.00 / fileTotal)))
                .averageProcessTime(TimeConvert(averageProcessTime1, null))
                .build();
    }

    /**
     * ms-s-min-时间处理
     * @param time 毫秒时间
     * @param type 指定转换类型1：s；2：min
     * @return
     */
    private String TimeConvert(BigDecimal time, Integer type) {
        BigDecimal convertTime = time;
        String convertTimeString = time + "ms";
        // 自动转换
        if (null == type) {
            // ms-s，大于1000毫秒转换
            if (convertTime.compareTo(new BigDecimal("1000")) >= 0) {
                convertTimeString = convertTime.divide(new BigDecimal("1000"), 2, BigDecimal.ROUND_HALF_UP) + "s";
                // s-min，大于60秒转换
                if (convertTime.compareTo(new BigDecimal("60")) >= 0) {
                    convertTimeString = convertTime.divide(new BigDecimal("60"), 2, BigDecimal.ROUND_HALF_UP) + "min";
                }
            }
        } else if (type == 1) {
            // 转换为s
            convertTimeString = convertTime.divide(new BigDecimal("1000"), 2, BigDecimal.ROUND_HALF_UP) + "s";
        } else if (type == 2) {
            // 转换为min
            convertTime = convertTime.divide(new BigDecimal("1000"), 2, BigDecimal.ROUND_HALF_UP);
            convertTimeString = convertTime.divide(new BigDecimal("60"), 2, BigDecimal.ROUND_HALF_UP) + "min";
        }
        return convertTimeString;
    }

    @Override
    @Cacheable(value = RedisConstantServer.DASHBOARD_SUCCESS_DATA, cacheManager = "myCacheManager", key = "T(String).valueOf(#queryDTO.toString())", unless = "#result == null")
    public List<DashboardDetailedDataDTO> getSuccessfulRequestDetailedData(DashboardQueryDTO queryDTO) {
        LambdaQueryWrapper<BackgroundConvertData> queryWrapper = getDashboardQueryWrapper(queryDTO);
        // 添加select 查询参数
        queryWrapper.select(BackgroundConvertData::getFileSuccessTotal, BackgroundConvertData::getSyncTime);
        // 查询数据
        List<BackgroundConvertData> backgroundConvertData = this.baseMapper.selectList(queryWrapper);

        LocalDateTime nowTime = getLocalDataTime();
        // 返回数据集合
        List<DashboardDetailedDataDTO> dataDTOS = new ArrayList<>();

        long successTotal = 0;
        // 判断查询类型
        switch (queryDTO.getQueryType()) {
            case 0:
                if (queryDTO.getDays() == 1) {
                    for (int i = 0; i < 24; i++) {
                        // 减去 i 小时时间
                        LocalDateTime minus = nowTime.minus(i, ChronoUnit.HOURS);
                        successTotal = backgroundConvertData.stream().filter(convertData -> localDataCompareTo(convertData.getSyncTime(), minus)
                                && !localDataCompareTo(convertData.getSyncTime(), minus.plusHours(1)))
                                .collect(Collectors.summarizingLong(BackgroundConvertData::getFileSuccessTotal)).getSum();

                        dataDTOS.add(DashboardDetailedDataDTO.builder()
                                .xData(minus.toString())
                                .yData(successTotal + "").build());
                    }
                } else {
                    for (int i = 0; i < queryDTO.getDays(); i++) {
                        // 减去 i 天时间
                        LocalDate minus = nowTime.minus(i, ChronoUnit.DAYS).toLocalDate();
                        successTotal = backgroundConvertData.stream().filter(convertData -> localDataCompareTo(convertData.getSyncTime(), minus.atStartOfDay())
                                && !localDataCompareTo(convertData.getSyncTime(), minus.plusDays(1).atStartOfDay()))
                                .collect(Collectors.summarizingLong(BackgroundConvertData::getFileSuccessTotal)).getSum();

                        dataDTOS.add(DashboardDetailedDataDTO.builder()
                                .xData(minus.toString())
                                .yData(successTotal + "").build());
//                        successTotal = 0;
                    }
                }
                break;
            case 1:
                LocalDate startDateTime = queryDTO.getStartDateTime();
                LocalDate endDateTime = queryDTO.getEndDateTime();
                long days = endDateTime.toEpochDay() - startDateTime.toEpochDay()  +1 ; //相差的天数
                if (days == 1L) {
                    endDateTime = endDateTime.plusDays(1L);
                    for (int i = 1; i <= 24; i++) {
                        // 减去 i 小时时间

                        LocalDateTime minus = endDateTime.atStartOfDay().minus(i, ChronoUnit.HOURS);

                        successTotal = backgroundConvertData.stream().filter(convertData -> localDataCompareTo(convertData.getSyncTime(), minus)
                                && !localDataCompareTo(convertData.getSyncTime(), minus.plus(1L, ChronoUnit.HOURS)))
                                .collect(Collectors.summarizingLong(BackgroundConvertData::getFileSuccessTotal)).getSum();

                        dataDTOS.add(DashboardDetailedDataDTO.builder()
                                .xData(minus.toString())
                                .yData(successTotal + "").build());
//                        successTotal = 0;
                    }
                } else {
                    for (int i = 0; i < days; i++) {
                        // 减去 i 天时间
                        LocalDate minus = endDateTime.minus(i, ChronoUnit.DAYS);

                        successTotal = backgroundConvertData.stream().filter(convertData -> localDataCompareTo(convertData.getSyncTime(), minus.atStartOfDay())
                                && !localDataCompareTo(convertData.getSyncTime(), minus.plusDays(1).atStartOfDay()))
                                .collect(Collectors.summarizingLong(BackgroundConvertData::getFileSuccessTotal)).getSum();
                        dataDTOS.add(DashboardDetailedDataDTO.builder()
                                .xData(minus.toString())
                                .yData(successTotal + "").build());
//                        successTotal = 0;
                    }
                }
                break;
            default:
                throw new CommonException(ResponseEnum.SYSTEM_ERROR);
        }
        return dataDTOS;
    }

    @Override
    @Cacheable(value = RedisConstantServer.DASHBOARD_FAIL_DATA, cacheManager = "myCacheManager", key = "T(String).valueOf(#queryDTO.toString())", unless = "#result == null")
    public List<DashboardDetailedDataDTO> getErrorRequestDetailedData(DashboardQueryDTO queryDTO) {
        LambdaQueryWrapper<BackgroundConvertData> queryWrapper = getDashboardQueryWrapper(queryDTO);
        // 添加select 查询参数
        queryWrapper.select(BackgroundConvertData::getFileFailedTotal, BackgroundConvertData::getSyncTime);
        // 查询数据
        List<BackgroundConvertData> backgroundConvertData = this.baseMapper.selectList(queryWrapper);
        // 判断查询类型
        LocalDateTime nowTime = getLocalDataTime();
        // 返回数据集合
        List<DashboardDetailedDataDTO> dataDTOS = new ArrayList<>();

        long fileFailedTotal = 0;
        switch (queryDTO.getQueryType()) {
            case 0:
                if (queryDTO.getDays() == 1) {
                    for (int i = 0; i < 24; i++) {
                        // 减去 i 小时时间
                        LocalDateTime minus = nowTime.minus(i, ChronoUnit.HOURS);

                        fileFailedTotal = backgroundConvertData.stream().filter(convertData -> localDataCompareTo(convertData.getSyncTime(), minus)
                                && !localDataCompareTo(convertData.getSyncTime(), minus.plusHours(1)))
                                .collect(Collectors.summarizingLong(BackgroundConvertData::getFileFailedTotal)).getSum();

                        dataDTOS.add(DashboardDetailedDataDTO.builder()
                                .xData(minus.toString())
                                .yData(fileFailedTotal + "").build());
//                        fileFailedTotal = 0;
                    }
                } else {
                    for (int i = 0; i < queryDTO.getDays(); i++) {
                        // 减去 i 天时间
                        LocalDate minus = nowTime.minus(i, ChronoUnit.DAYS).toLocalDate();

                        fileFailedTotal = backgroundConvertData.stream().filter(convertData -> localDataCompareTo(convertData.getSyncTime(), minus.atStartOfDay())
                                && !localDataCompareTo(convertData.getSyncTime(), minus.plusDays(1).atStartOfDay()))
                                .collect(Collectors.summarizingLong(BackgroundConvertData::getFileFailedTotal)).getSum();

                        dataDTOS.add(DashboardDetailedDataDTO.builder()
                                .xData(minus.toString())
                                .yData(fileFailedTotal + "").build());
//                        fileFailedTotal = 0;
                    }
                }
                break;
            case 1:
                LocalDate startDateTime = queryDTO.getStartDateTime();
                LocalDate endDateTime = queryDTO.getEndDateTime();
                long days = endDateTime.toEpochDay() - startDateTime.toEpochDay()  +1 ; //相差的天数
                if (days == 1L) {
                    endDateTime = endDateTime.plusDays(1L);
                    for (int i = 1; i <= 24; i++) {
                        // 减去 i 小时时间

                        LocalDateTime minus = endDateTime.atStartOfDay().minus(i, ChronoUnit.HOURS);
                        fileFailedTotal = backgroundConvertData.stream().filter(convertData -> localDataCompareTo(convertData.getSyncTime(), minus)
                                && !localDataCompareTo(convertData.getSyncTime(), minus.plus(1L, ChronoUnit.HOURS)))
                                .collect(Collectors.summarizingLong(BackgroundConvertData::getFileFailedTotal)).getSum();
                        dataDTOS.add(DashboardDetailedDataDTO.builder()
                                .xData(minus.toString())
                                .yData(fileFailedTotal + "").build());
//                        fileFailedTotal = 0;
                    }
                } else {
                    for (int i = 0; i < days; i++) {
                        // 减去 i 天时间
                        LocalDate minus = endDateTime.minus(i, ChronoUnit.DAYS);

                        fileFailedTotal = backgroundConvertData.stream().filter(convertData -> localDataCompareTo(convertData.getSyncTime(), minus.atStartOfDay())
                                && !localDataCompareTo(convertData.getSyncTime(), minus.plusDays(1).atStartOfDay()))
                                .collect(Collectors.summarizingLong(BackgroundConvertData::getFileFailedTotal)).getSum();
                        dataDTOS.add(DashboardDetailedDataDTO.builder()
                                .xData(minus.toString())
                                .yData(fileFailedTotal + "").build());
//                        fileFailedTotal = 0;
                    }
                }
                break;
            default:
                throw new CommonException(ResponseEnum.SYSTEM_ERROR);
        }
        return dataDTOS;
    }

    @Override
    @Cacheable(value = RedisConstantServer.DASHBOARD_ERROR_RATIO_DATA, cacheManager = "myCacheManager", key = "T(String).valueOf(#queryDTO.toString())", unless = "#result == null")
    public List<DashboardDetailedDataDTO> getErrorRatioDetailedData(DashboardQueryDTO queryDTO) {
        List<DashboardDetailedDataDTO> successfulRequestDetailedData = this.getSuccessfulRequestDetailedData(queryDTO);
        List<DashboardDetailedDataDTO> errorRequestDetailedData = this.getErrorRequestDetailedData(queryDTO);
        List<DashboardDetailedDataDTO> dataDTOS = new ArrayList<>();
        for (int i = 0; i < successfulRequestDetailedData.size(); i++) {

            int fileTotal = Integer.parseInt(successfulRequestDetailedData.get(i).getYData()) + Integer.parseInt(errorRequestDetailedData.get(i).getYData());

            dataDTOS.add(DashboardDetailedDataDTO.builder()
                    .xData(successfulRequestDetailedData.get(i).getXData())
                    .yData(fileTotal == 0 ? "0.00%" : new BigDecimal(String.format("%.2f", Integer.parseInt(errorRequestDetailedData.get(i).getYData()) * 100.00
                            / fileTotal)).toString() + "%")
                    .build());
        }
        return dataDTOS;
    }

    @Override
    @Cacheable(value = RedisConstantServer.DASHBOARD_AVERAGE_PROCESS_DATA, cacheManager = "myCacheManager", key = "T(String).valueOf(#queryDTO.toString())", unless = "#result == null")
    public List<DashboardDetailedDataDTO> getAverageProcessTimeDetailedData(DashboardQueryDTO queryDTO) {
        LambdaQueryWrapper<BackgroundConvertData> queryWrapper = getDashboardQueryWrapper(queryDTO);
        // 添加select 查询参数
        queryWrapper.select(BackgroundConvertData::getFileConvertTime, BackgroundConvertData::getSyncTime);
        // 查询数据
        List<BackgroundConvertData> backgroundConvertData = this.baseMapper.selectList(queryWrapper);
        // 判断查询类型
        LocalDateTime nowTime = getLocalDataTime();
        // 返回数据集合
        List<DashboardDetailedDataDTO> dataDTOS = new ArrayList<>();

        BigDecimal convertTime = BigDecimal.ZERO;
        switch (queryDTO.getQueryType()) {
            case 0:
                if (queryDTO.getDays() == 1) {
                    for (int i = 0; i < 24; i++) {
                        // 减去 i 小时时间
                        LocalDateTime minus = nowTime.minus(i, ChronoUnit.HOURS);
                        long count = backgroundConvertData.stream().filter(convertData -> localDataCompareTo(convertData.getSyncTime(), minus)
                                && !localDataCompareTo(convertData.getSyncTime(), minus.plusHours(1)) && convertData.getFileConvertTime() != null &&
                                convertData.getFileConvertTime().compareTo(BigDecimal.ZERO) > 0).count();
                        if (count <= 0) {
                            convertTime = BigDecimal.ZERO;
                        } else {
                            convertTime = backgroundConvertData.stream().filter(convertData -> localDataCompareTo(convertData.getSyncTime(), minus)
                                    && !localDataCompareTo(convertData.getSyncTime(), minus.plusHours(1)) && convertData.getFileConvertTime() != null &&
                                    convertData.getFileConvertTime().compareTo(BigDecimal.ZERO) > 0)
                                    .map(BackgroundConvertData::getFileConvertTime).reduce(BigDecimal.ZERO, BigDecimal::add)
                                    .divide(BigDecimal.valueOf(count), 2, BigDecimal.ROUND_HALF_UP);
                        }
                        dataDTOS.add(DashboardDetailedDataDTO.builder()
                                .xData(minus.toString())
                                .yData(convertTime.stripTrailingZeros().toPlainString()).build());
                    }
                } else {
                    for (int i = 0; i < queryDTO.getDays(); i++) {
                        // 减去 i 天时间
                        LocalDate minus = nowTime.minus(i, ChronoUnit.DAYS).toLocalDate();
                        long count = backgroundConvertData.stream().filter(convertData -> localDataCompareTo(convertData.getSyncTime(), minus.atStartOfDay())
                                && !localDataCompareTo(convertData.getSyncTime(), minus.plusDays(1).atStartOfDay()) && convertData.getFileConvertTime() != null &&
                                convertData.getFileConvertTime().compareTo(BigDecimal.ZERO) > 0).count();
                        if (count <= 0) {
                            convertTime = BigDecimal.ZERO;
                        } else {
                            convertTime = backgroundConvertData.stream().filter(convertData -> localDataCompareTo(convertData.getSyncTime(), minus.atStartOfDay())
                                    && !localDataCompareTo(convertData.getSyncTime(), minus.plusDays(1).atStartOfDay()) && convertData.getFileConvertTime() != null &&
                                    convertData.getFileConvertTime().compareTo(BigDecimal.ZERO) > 0)
                                    .map(BackgroundConvertData::getFileConvertTime).reduce(BigDecimal.ZERO, BigDecimal::add)
                                    .divide(BigDecimal.valueOf(count), 2, BigDecimal.ROUND_HALF_UP);
                        }
                        dataDTOS.add(DashboardDetailedDataDTO.builder()
                                .xData(minus.toString())
                                .yData(convertTime.stripTrailingZeros().toPlainString()).build());
                    }
                }
                break;
            case 1:
                LocalDate startDateTime = queryDTO.getStartDateTime();
                LocalDate endDateTime = queryDTO.getEndDateTime();
                long days = endDateTime.toEpochDay() - startDateTime.toEpochDay()  +1 ; //相差的天数
                if (days == 1L) {
                    endDateTime = endDateTime.plusDays(1L);
                    for (int i = 1; i <= 24; i++) {
                        // 减去 i 小时时间
                        LocalDateTime minus = endDateTime.atStartOfDay().minus(i, ChronoUnit.HOURS);
                        long count = backgroundConvertData.stream().filter(convertData -> localDataCompareTo(convertData.getSyncTime(), minus)
                                && !localDataCompareTo(convertData.getSyncTime(), minus.plus(1L, ChronoUnit.HOURS)) && convertData.getFileConvertTime() != null &&
                                convertData.getFileConvertTime().compareTo(BigDecimal.ZERO) > 0).count();
                        if (count <= 0) {
                            convertTime = BigDecimal.ZERO;
                        } else {
                            convertTime = backgroundConvertData.stream().filter(convertData -> localDataCompareTo(convertData.getSyncTime(), minus)
                                    && !localDataCompareTo(convertData.getSyncTime(), minus.plus(1L, ChronoUnit.HOURS)) && convertData.getFileConvertTime() != null &&
                                    convertData.getFileConvertTime().compareTo(BigDecimal.ZERO) > 0)
                                    .map(BackgroundConvertData::getFileConvertTime).reduce(BigDecimal.ZERO, BigDecimal::add)
                                    .divide(BigDecimal.valueOf(count), 2, BigDecimal.ROUND_HALF_UP);
                        }
                        dataDTOS.add(DashboardDetailedDataDTO.builder()
                                .xData(minus.toString())
                                .yData(convertTime.stripTrailingZeros().toPlainString()).build());
                    }
                } else {
                    for (int i = 0; i < days; i++) {
                        // 减去 i 天时间
                        LocalDate minus = endDateTime.minus(i, ChronoUnit.DAYS);
                        long count = backgroundConvertData.stream().filter(convertData -> localDataCompareTo(convertData.getSyncTime(), minus.atStartOfDay())
                                && !localDataCompareTo(convertData.getSyncTime(), minus.plusDays(1).atStartOfDay()) && convertData.getFileConvertTime() != null &&
                                convertData.getFileConvertTime().compareTo(BigDecimal.ZERO) > 0).count();
                        if (count <= 0) {
                            convertTime = BigDecimal.ZERO;
                        } else {
                            convertTime = backgroundConvertData.stream().filter(convertData -> localDataCompareTo(convertData.getSyncTime(), minus.atStartOfDay())
                                    && !localDataCompareTo(convertData.getSyncTime(), minus.plusDays(1).atStartOfDay()) && convertData.getFileConvertTime() != null &&
                                    convertData.getFileConvertTime().compareTo(BigDecimal.ZERO) > 0)
                                    .map(BackgroundConvertData::getFileConvertTime).reduce(BigDecimal.ZERO, BigDecimal::add)
                                    .divide(BigDecimal.valueOf(count), 2, BigDecimal.ROUND_HALF_UP);
                        }
                        dataDTOS.add(DashboardDetailedDataDTO.builder()
                                .xData(minus.toString())
                                .yData(convertTime.stripTrailingZeros().toPlainString()).build());
                    }
                }
                break;
            default:
                throw new CommonException(ResponseEnum.SYSTEM_ERROR);
        }

        // 计算平均值，确定单位
        BigDecimal average = BigDecimal.ZERO;
        for (DashboardDetailedDataDTO dataDTO : dataDTOS) {
            average = average.add(new BigDecimal(dataDTO.getYData()));
        }
        average = average.divide(new BigDecimal(dataDTOS.size()), 2, BigDecimal.ROUND_HALF_UP);
        Integer flag = null;
        // 平均值大于秒
        if (average.compareTo(new BigDecimal("1000")) >= 0) {
            flag = 1;
        }
        // 平均值大于分钟
        if (average.compareTo(new BigDecimal("60000")) >= 0) {
            flag = 2;
        }
        // 转换
        if (null != flag) {
            for (DashboardDetailedDataDTO dataDTO : dataDTOS) {
                dataDTO.setYData(TimeConvert(new BigDecimal(dataDTO.getYData()), flag));
            }
        } else {
            for (DashboardDetailedDataDTO dataDTO : dataDTOS) {
                dataDTO.setYData(dataDTO.getYData() + "ms");
            }
        }
        return dataDTOS;
    }

    @Override
    public void dashboardExport(DashboardQueryDTO queryDTO, HttpServletResponse response) {
        List<DashboardDetailedDataDTO> successfulRequestDetailedData = this.getSuccessfulRequestDetailedData(queryDTO);
        List<DashboardDetailedDataDTO> errorRequestDetailedData = this.getErrorRequestDetailedData(queryDTO);
        List<DashboardDetailedDataDTO> errorRatioDetailedData = this.getErrorRatioDetailedData(queryDTO);
        List<DashboardDetailedDataDTO> averageProcessTimeDetailedData = this.getAverageProcessTimeDetailedData(queryDTO);
        List<DashboardExportObject> dashboardExportObjects = new ArrayList<>();
        boolean hourOrDate = (queryDTO.getQueryType() == 0 && queryDTO.getDays() == 1) || (queryDTO.getQueryType() == 1 && queryDTO.getStartDateTime().equals(queryDTO.getEndDateTime()));
        String time;
        for (int i = 0; i < successfulRequestDetailedData.size(); i++) {
            time = successfulRequestDetailedData.get(i).getXData();
            if (hourOrDate) {
                time = time.replaceFirst("T", " ");
            }
            dashboardExportObjects.add(DashboardExportObject.builder()
                    .time(time)
                    .successfulRequests(successfulRequestDetailedData.get(i).getYData())
                    .errorRequests(errorRequestDetailedData.get(i).getYData())
                    .errorRatio(errorRatioDetailedData.get(i).getYData())
                    .averageProcessTime(averageProcessTimeDetailedData.get(i).getYData() + "ms")
                    .build());
        }

        response.setContentType("application/vnd.openxmlformats-officedocument.spreadsheetml.sheet");
        response.setCharacterEncoding("utf-8");
        List<List<String>> listHand = new ArrayList<>();
        List<String> head0 = new ArrayList<>();
        // 判断是Hour还是Date
        head0.add(hourOrDate ? "Hour" : "Date");
        List<String> head1 = new ArrayList<>();
        head1.add("Succesful Requests");
        List<String> head2 = new ArrayList<>();
        head2.add("Error Requests");
        List<String> head3 = new ArrayList<>();
        head3.add("Error Ratio");
        List<String> head4 = new ArrayList<>();
        head4.add("Average Process Time");
        listHand.add(head0);
        listHand.add(head1);
        listHand.add(head2);
        listHand.add(head3);
        listHand.add(head4);
        try {
            // 这里URLEncoder.encode可以防止中文乱码 当然和easyexcel没有关系
            String fileName = URLEncoder.encode("Dashboard表单", "UTF-8").replaceAll("\\+", "%20");
            response.setHeader("Content-disposition", "attachment;filename*=utf-8''" + fileName + ".xlsx");
            EasyExcel.write(response.getOutputStream())
                    .head(listHand)
                    .registerWriteHandler(new ExcelWidthStyleStrategy())
                    .autoCloseStream(Boolean.FALSE)
                    .registerWriteHandler(new LongestMatchColumnWidthStyleStrategy())
                    .sheet("Dashboard")
                    .doWrite(dashboardExportObjects);
        } catch (Exception e) {
            // 重置response
            response.reset();
            response.setContentType("application/json");
            response.setCharacterEncoding("utf-8");
            Map<String, String> map = MapUtils.newHashMap();
            map.put("status", "failure");
            map.put("message", "下载文件失败" + e.getMessage());
            try {
                response.getWriter().println(JSON.toJSONString(map));
            } catch (IOException ioException) {
                ioException.printStackTrace();
            }
        }
    }

    @Override
    @CacheEvict(cacheNames = RedisConstantServer.DASHBOARD + "*", cacheManager = "myCacheManager", allEntries = true)
    public void dashboardDataSync(List<DashboardStatisticsDTO> list) {
        assert list != null : "Dashboard同步数据为空";
        list.forEach(item ->{
            item.setUserId(userService.getUserId(item.getTenantId()));
        });
        final String DASHBOARD_SYNC_KEY = "DashboardSyncKey:" + list.get(0).getSyncTime();
        Boolean b = redisTemplate.hasKey(DASHBOARD_SYNC_KEY);
        assert b != null;
        if (!b) {
            log.info("开始数据同步");
            redisTemplate.opsForValue().setIfAbsent(DASHBOARD_SYNC_KEY, "", 60, TimeUnit.MINUTES);
            try {
                this.baseMapper.insertList(list);
            } catch (Exception e) {
                redisTemplate.delete(DASHBOARD_SYNC_KEY);
                throw e;
            }
        } else {
            log.error("当前时间节点已同步过：{}" + DASHBOARD_SYNC_KEY);
        }
    }

    @Override
    public void refresh(Long tenantId) {
        String cacheKey = ":DashboardQueryDTO(tenantId=" +
                tenantId +
                ",*";
        String analysisData = RedisConstantServer.DASHBOARD_ANALYSIS_DATA + cacheKey;
        String successData = RedisConstantServer.DASHBOARD_SUCCESS_DATA + cacheKey;
        String failData = RedisConstantServer.DASHBOARD_FAIL_DATA + cacheKey;
        String errorRatioData = RedisConstantServer.DASHBOARD_ERROR_RATIO_DATA + cacheKey;
        String averageProcessData = RedisConstantServer.DASHBOARD_AVERAGE_PROCESS_DATA + cacheKey;

        redisTemplate.delete(Objects.requireNonNull(redisTemplate.keys(analysisData)));
        redisTemplate.delete(Objects.requireNonNull(redisTemplate.keys(successData)));
        redisTemplate.delete(Objects.requireNonNull(redisTemplate.keys(failData)));
        redisTemplate.delete(Objects.requireNonNull(redisTemplate.keys(errorRatioData)));
        redisTemplate.delete(Objects.requireNonNull(redisTemplate.keys(averageProcessData)));
        log.info("删除缓存成功，tenantId：{}", tenantId);
        // TODO 强制刷新去同步转档数据库数据
    }


    /**
     * 获取Dashboard 查询使用的LambdaQueryWrapper
     *
     * @param queryDTO 查询参数
     * @return Dashboard 查询使用的LambdaQueryWrapper
     */
    private LambdaQueryWrapper<BackgroundConvertData> getDashboardQueryWrapper(DashboardQueryDTO queryDTO) {
        LambdaQueryWrapper<BackgroundConvertData> queryWrapper = new LambdaQueryWrapper<>();
        // 查询类型
        Integer queryType = queryDTO.getQueryType();
        log.info("本次查询AnalysisData参数：{}", queryDTO);
        /* 如果时间范围查询为系统定义 */
        if (queryType == 0) {
            Integer days = queryDTO.getDays();
            if (days > 30) {
                days = 30;
                queryDTO.setDays(30);
            } else if (days < 1) {
                days = 1;
                queryDTO.setDays(1);
            }
            LocalDateTime nowDateTime = LocalDateTime.now();
            LocalDateTime minus = nowDateTime.minus(days, ChronoUnit.DAYS);
            queryWrapper.ge(BackgroundConvertData::getSyncTime, minus);
        } else if (queryType == 1) {
            LocalDate startDateTime = queryDTO.getStartDateTime();
            LocalDate endDateTime = queryDTO.getEndDateTime();

            if (startDateTime.isAfter(endDateTime)) {
                // 查询参数异常 开始时间大于结束时间
                throw new CommonException(BackgroundErrorEnum.QUEUE_PARAMETER_ERROR);
            } else if (startDateTime.equals(endDateTime)) {
                // 开始时间等于结束时间
                endDateTime = endDateTime.plusDays(1);
            } else if (ChronoUnit.DAYS.between(startDateTime, endDateTime) > 30) {
                // 开始时间和结束时间大于30天
                startDateTime = endDateTime.minus(30, ChronoUnit.DAYS);
                queryDTO.setStartDateTime(endDateTime.minus(30, ChronoUnit.DAYS));
            }
            queryWrapper.ge(BackgroundConvertData::getSyncTime, startDateTime)
                    .le(BackgroundConvertData::getSyncTime, endDateTime);
        } else {
            throw new CommonException(BackgroundErrorEnum.ERROR_INNER);
        }
        // 如果条件 projectId不为空则添加查询
        if (queryDTO.getProjectId() != null) {
            queryWrapper.eq(BackgroundConvertData::getProjectId, queryDTO.getProjectId());
        }
        // 如果条件 toolId不为空则添加查询
        if (queryDTO.getToolId() != null) {
            queryWrapper.eq(BackgroundConvertData::getToolId, queryDTO.getToolId());
        }
        queryWrapper.eq(BackgroundConvertData::getTenantId, queryDTO.getTenantId());
        return queryWrapper;
    }

    /**
     * 获取整数小时时间
     */
    private static LocalDateTime getLocalDataTime() {
        LocalDateTime nowTime = LocalDateTime.now();
        LocalDate nowData = LocalDate.now();
        return LocalDateTime.of(nowData, LocalTime.of(nowTime.getHour(), 0));
    }

    private static boolean localDataCompareTo(LocalDateTime dateTime1, LocalDateTime dateTime2) {
        return dateTime1.compareTo(dateTime2) >= 0;
    }

}
